"""
This module contains the websocket for TSETMC
"""
import asyncio
import json
from dataclasses import dataclass
import logging
from typing import Callable, Awaitable
from threading import Lock
from websockets.server import serve
from websockets.sync.client import ClientConnection
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from tse_utils.models.instrument import Instrument
from tsetmc_pusher.server.repository import MarketRealtimeData
from tsetmc_pusher.server.timing import sleep_until, MARKET_END_TIME


@dataclass
class InstrumentChannel:
    """Holds essential channels for each instrument"""

    isin: str = None
    trade_subscribers: set[ClientConnection] = None
    orderbook_subscribers: set[ClientConnection] = None
    clienttype_subscribers: set[ClientConnection] = None

    def __init__(self, isin: str):
        self.isin = isin
        self.trade_subscribers = set()
        self.orderbook_subscribers = set()
        self.clienttype_subscribers = set()

    def __repr__(self) -> str:
        return f"{self.isin}: {[x.id for x in self.orderbook_subscribers]}"


def subscribe_trade(client: ClientConnection, instrument_channel: InstrumentChannel):
    """Subscribe to instrument's trade data"""
    instrument_channel.trade_subscribers.add(client)


def subscribe_orderbook(
    client: ClientConnection, instrument_channel: InstrumentChannel
):
    """Subscribe to instrument's orderbook data"""
    instrument_channel.orderbook_subscribers.add(client)


def subscribe_clienttype(
    client: ClientConnection, instrument_channel: InstrumentChannel
):
    """Subscribe to instrument's clienttype data"""
    instrument_channel.clienttype_subscribers.add(client)


def subscribe_all(client: ClientConnection, instrument_channel: InstrumentChannel):
    """Subscribe to all instrument's data"""
    subscribe_trade(client, instrument_channel)
    subscribe_orderbook(client, instrument_channel)
    subscribe_clienttype(client, instrument_channel)


def unsubscribe_trade(client: ClientConnection, instrument_channel: InstrumentChannel):
    """Unsubscribe from instrument's trade data"""
    try:
        instrument_channel.trade_subscribers.remove(client)
    except KeyError:
        pass


def unsubscribe_orderbook(
    client: ClientConnection, instrument_channel: InstrumentChannel
):
    """Unsubscribe from instrument's orderbook data"""
    try:
        instrument_channel.orderbook_subscribers.remove(client)
    except KeyError:
        pass


def unsubscribe_clienttype(
    client: ClientConnection, instrument_channel: InstrumentChannel
):
    """Unsubscribe from instrument's clienttype data"""
    try:
        instrument_channel.clienttype_subscribers.remove(client)
    except KeyError:
        pass


def unsubscribe_all(client: ClientConnection, instrument_channel: InstrumentChannel):
    """Unsubscribe from all instrument's data"""
    unsubscribe_trade(client, instrument_channel)
    unsubscribe_orderbook(client, instrument_channel)
    unsubscribe_clienttype(client, instrument_channel)


def instrument_data_trade(instrument: Instrument) -> list:
    """Convert instrument's trade data for websocket transfer"""
    ltd = instrument.intraday_trade_candle.last_trade_datetime
    return {
        "trade": [
            instrument.intraday_trade_candle.close_price,
            instrument.intraday_trade_candle.last_price,
            str(ltd),
            instrument.intraday_trade_candle.max_price,
            instrument.intraday_trade_candle.min_price,
            instrument.intraday_trade_candle.open_price,
            instrument.intraday_trade_candle.previous_price,
            instrument.intraday_trade_candle.trade_num,
            instrument.intraday_trade_candle.trade_value,
            instrument.intraday_trade_candle.trade_volume,
        ]
    }


def instrument_data_orderbook_specific_rows(
    instrument: Instrument, rows: list[int]
) -> list:
    """Convert instrument's orderbook data for websocket transfer"""
    return {
        "orderbook": [
            [
                rn,
                x.demand.num,
                x.demand.price,
                x.demand.volume,
                x.supply.num,
                x.supply.price,
                x.supply.volume,
            ]
            for rn, x in enumerate(instrument.orderbook.rows)
            if rn in rows
        ]
    }


def instrument_data_orderbook(instrument: Instrument) -> list:
    """Convert instrument's orderbook data for websocket transfer"""
    return {
        "orderbook": [
            [
                rn,
                x.demand.num,
                x.demand.price,
                x.demand.volume,
                x.supply.num,
                x.supply.price,
                x.supply.volume,
            ]
            for rn, x in enumerate(instrument.orderbook.rows)
        ]
    }


def instrument_data_clienttype(instrument: Instrument) -> list[int]:
    """Convert instrument's clienttype data for websocket transfer"""
    return {
        "clienttype": [
            instrument.client_type.legal.buy.num,
            instrument.client_type.legal.buy.volume,
            instrument.client_type.legal.sell.num,
            instrument.client_type.legal.sell.volume,
            instrument.client_type.natural.buy.num,
            instrument.client_type.natural.buy.volume,
            instrument.client_type.natural.sell.num,
            instrument.client_type.natural.sell.volume,
        ]
    }


def instrument_data_thresholds(instrument: Instrument) -> list[int]:
    """Convert instrument's price thresholds data for websocket transfer"""
    return {
        "thresholds": [
            instrument.order_limitations.max_price,
            instrument.order_limitations.min_price,
        ]
    }


def instrument_data_all(instrument: Instrument) -> dict[str, list]:
    """Convert all instrument's data for websocket transfer"""
    return (
        instrument_data_thresholds(instrument)
        | instrument_data_trade(instrument)
        | instrument_data_orderbook(instrument)
        | instrument_data_clienttype(instrument)
    )


class TsetmcWebsocket:
    """Holds the websocket for TSETMC"""

    _LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        market_realtime_data: MarketRealtimeData,
        websocket_host: str,
        websocket_port: int,
    ):
        self.market_realtime_data: MarketRealtimeData = market_realtime_data
        self.websocket_host: str = websocket_host
        self.websocket_port: int = websocket_port
        self.__channels: list[InstrumentChannel] = []
        self.__channels_lock = Lock()
        self.set_market_realtime_data_pushers()

    def set_market_realtime_data_pushers(self) -> None:
        """Sets the pusher methods for market realtime data"""
        self.market_realtime_data.pusher_trade_data = self.pusher_trade_data
        self.market_realtime_data.pusher_orderbook_data = self.pusher_orderbook_data
        self.market_realtime_data.pusher_clienttype_data = self.pusher_clienttype_data

    async def pusher_trade_data(
        self, instruments: list[Instrument]
    ) -> Callable[[list[Instrument]], Awaitable[None]]:
        """Returns the pusher_trade_data to override in repo"""
        for instrument in instruments:
            with self.__channels_lock:
                channel = next(
                    (
                        x
                        for x in self.__channels
                        if x.isin == instrument.identification.isin
                    ),
                    None,
                )
                if channel and channel.trade_subscribers:
                    await self.broadcast(
                        channel.trade_subscribers,
                        json.dumps({channel.isin: instrument_data_trade(instrument)}),
                    )

    async def pusher_orderbook_data(
        self, instruments: list[Instrument]
    ) -> Callable[[list[tuple[Instrument, list[int]]]], Awaitable[None]]:
        """Returns the pusher_orderbook_data to override in repo"""
        for instrument, rows in instruments:
            with self.__channels_lock:
                channel = next(
                    (
                        x
                        for x in self.__channels
                        if x.isin == instrument.identification.isin
                    ),
                    None,
                )
                if channel and channel.orderbook_subscribers:
                    await self.broadcast(
                        channel.orderbook_subscribers,
                        json.dumps(
                            {
                                channel.isin: instrument_data_orderbook_specific_rows(
                                    instrument, rows
                                )
                            }
                        ),
                    )

    async def pusher_clienttype_data(
        self, instruments: list[Instrument]
    ) -> Callable[[list[Instrument]], Awaitable[None]]:
        """Returns the pusher_clienttype_data to override in repo"""
        for instrument in instruments:
            with self.__channels_lock:
                channel = next(
                    (
                        x
                        for x in self.__channels
                        if x.isin == instrument.identification.isin
                    ),
                    None,
                )
                if channel and channel.clienttype_subscribers:
                    await self.broadcast(
                        channel.clienttype_subscribers,
                        json.dumps(
                            {channel.isin: instrument_data_clienttype(instrument)}
                        ),
                    )

    @classmethod
    async def broadcast(cls, clients: list[ClientConnection], message: str) -> None:
        """Broadcast a message to a bunch of users"""
        group = asyncio.gather(*[client.send(message) for client in clients])
        await asyncio.wait_for(group, timeout=None)

    async def handle_connection(self, client: ClientConnection) -> None:
        """Handles the clients' connections"""
        self._LOGGER.info("Connection opened to [%s]", client.id)
        try:
            async for message in client:
                self._LOGGER.info(
                    "Receieved message [%s] from [%s]", message, client.id
                )
                response = self.handle_connection_message(client, message)
                if response:
                    await client.send(json.dumps(response))
        except (ConnectionClosedError, ConnectionClosedOK):
            self._LOGGER.info("Connection closed to [%s]", client.id)
        finally:
            self._LOGGER.info("Removing [%s] from all channels", client.id)
            self.remove_from_channels(client)

    def remove_from_channels(self, client: ClientConnection) -> None:
        """Removes a client from all channels"""
        with self.__channels_lock:
            for channel in self.__channels:
                unsubscribe_all(client, channel)

    def handle_connection_message(self, client: ClientConnection, message: str) -> dict:
        """
        Handles a single message from client
        Standard message format is: <Action>.<Channel>.<Isin1>,<Isin2>,...
        For instance: 1.trade.IRO1FOLD0001,IRO1IKCO0001
        """
        acceptable_actions = ["0", "1"]
        acceptable_channels = ["all", "trade", "orderbook", "clienttype"]
        message_parts = message.split(".")
        if len(message_parts) != 3:
            self._LOGGER.error("Message [%s] has unacceptable format.", message)
            return None
        if message_parts[0] not in acceptable_actions:
            self._LOGGER.error("Action [%s] is not acceptable.", message_parts[0])
            return None
        if message_parts[1] not in acceptable_channels:
            self._LOGGER.error("Channel [%s] is not acceptable.", message_parts[1])
            return None
        if message_parts[2] == "*":
            instruments = self.market_realtime_data.get_all_instruments()
            isins = [x.identification.isin for x in instruments]
        else:
            isins = message_parts[2].split(",")
            fake_isin = next((x for x in isins if len(x) != 12), None)
            if fake_isin:
                self._LOGGER.error("Isin [%s] is not acceptable.", fake_isin)
                return None
            instruments = self.market_realtime_data.get_instruments(isins)
        channel_action_func = self.get_channel_action_func(
            message_parts[0], message_parts[1]
        )
        initial_data_func = self.get_initial_data_func(
            message_parts[0], message_parts[1]
        )
        initial_data = {}
        with self.__channels_lock:
            for counter, isin in enumerate(isins):
                channel = next((x for x in self.__channels if x.isin == isin), None)
                if not channel:
                    channel = InstrumentChannel(isin)
                    self.__channels.append(channel)
                    self._LOGGER.info("New channel for [%s]", isin)
                channel_action_func(client, channel)
                if instruments[counter]:
                    initial_data[isin] = initial_data_func(instruments[counter])
        return initial_data

    def get_channel_action_func(
        self, action: str, channel: str
    ) -> Callable[[ClientConnection, InstrumentChannel], None]:
        """Returns the function for handling the subscriptions"""
        return_values = {
            "1": {
                "all": subscribe_all,
                "trade": subscribe_trade,
                "orderbook": subscribe_orderbook,
                "clienttype": subscribe_clienttype,
            },
            "0": {
                "all": unsubscribe_all,
                "trade": unsubscribe_trade,
                "orderbook": unsubscribe_orderbook,
                "clienttype": unsubscribe_clienttype,
            },
        }
        return return_values[action][channel]

    def get_initial_data_func(
        self, action: str, channel: str
    ) -> Callable[[Instrument], None]:
        """Returns the method for getting the initial data after subscription"""
        return_values = {
            "1": {
                "all": instrument_data_all,
                "trade": instrument_data_trade,
                "orderbook": instrument_data_orderbook,
                "clienttype": instrument_data_clienttype,
            },
            "0": {
                "all": lambda x: None,
                "trade": lambda x: None,
                "orderbook": lambda x: None,
                "clienttype": lambda x: None,
            },
        }
        return return_values[action][channel]

    async def serve_websocket(self) -> None:
        """Serves the websocket for the project"""
        self._LOGGER.info("Serving has started.")
        async with serve(
            self.handle_connection, self.websocket_host, self.websocket_port
        ):
            await sleep_until(MARKET_END_TIME)
        self._LOGGER.info("Serving has ended.")



"""
This module contains the necessary codes for the TSETMC pusher's client.
"""
import json
import logging
import asyncio
from datetime import datetime
from websockets import client
from websockets.exceptions import ConnectionClosedError
from websockets.sync.client import ClientConnection
from tse_utils.models.instrument import Instrument


class TsetmcClient:
    """
The class used for connecting to the TSETMC pusher websocket \
and subscribe to its realtime data
    """

    _LOGGER: logging.Logger = logging.getLogger(__name__)
    _OPERATION_RECONNECT_WAIT: int = 5

    def __init__(
        self,
        websocket_host: str,
        websocket_port: int,
        subscribed_instruments: list[Instrument],
    ):
        self.websocket_host: str = websocket_host
        self.websocket_port: int = websocket_port
        self.websocket: ClientConnection = None
        self.subscribed_instruments: list[Instrument] = subscribed_instruments
        self.operation_flag: bool = False

    async def listen(self) -> None:
        """Listens to websocket updates"""
        while self.operation_flag:
            message = await self.websocket.recv()
            self._LOGGER.debug("Client received: %s", message)
            self.process_message(message=message)

    def process_message(self, message: str) -> None:
        """Processes a new message received from websocket"""
        message_js = json.loads(message)
        for isin, channels in message_js.items():
            instrument = next(
                x for x in self.subscribed_instruments if x.identification.isin == isin
            )
            for channel, data in channels.items():
                match channel:
                    case "thresholds":
                        self.__message_thresholds(instrument, data)
                    case "trade":
                        self.__message_trade(instrument, data)
                    case "orderbook":
                        self.__message_orderbook(instrument, data)
                    case "clienttype":
                        self.__message_clienttype(instrument, data)
                    case _:
                        self._LOGGER.fatal("Unknown message channel: %s", channel)

    def __message_thresholds(self, instrument: Instrument, data: list) -> None:
        """Handles a threshold update message"""
        instrument.order_limitations.max_price = int(data[0])
        instrument.order_limitations.min_price = int(data[1])

    def __message_trade(self, instrument: Instrument, data: list) -> None:
        """Handles a trade update message"""
        instrument.intraday_trade_candle.close_price = int(data[0])
        instrument.intraday_trade_candle.last_price = int(data[1])
        instrument.intraday_trade_candle.last_trade_datetime = datetime.fromisoformat(
            data[2]
        )
        instrument.intraday_trade_candle.max_price = int(data[3])
        instrument.intraday_trade_candle.min_price = int(data[4])
        instrument.intraday_trade_candle.open_price = int(data[5])
        instrument.intraday_trade_candle.previous_price = int(data[6])
        instrument.intraday_trade_candle.trade_num = int(data[7])
        instrument.intraday_trade_candle.trade_value = int(data[8])
        instrument.intraday_trade_candle.trade_volume = int(data[9])

    def __message_orderbook(self, instrument: Instrument, data: list) -> None:
        """Handles an orderbook update message"""
        for row in data:
            rn = int(row[0])
            instrument.orderbook.rows[rn].demand.num = int(row[1])
            instrument.orderbook.rows[rn].demand.price = int(row[2])
            instrument.orderbook.rows[rn].demand.volume = int(row[3])
            instrument.orderbook.rows[rn].supply.num = int(row[4])
            instrument.orderbook.rows[rn].supply.price = int(row[5])
            instrument.orderbook.rows[rn].supply.volume = int(row[6])

    def __message_clienttype(self, instrument: Instrument, data: list) -> None:
        """Handles an orderbook update message"""
        instrument.client_type.legal.buy.num = int(data[0])
        instrument.client_type.legal.buy.volume = int(data[1])
        instrument.client_type.legal.sell.num = int(data[2])
        instrument.client_type.legal.sell.volume = int(data[3])
        instrument.client_type.natural.buy.num = int(data[4])
        instrument.client_type.natural.buy.volume = int(data[5])
        instrument.client_type.natural.sell.num = int(data[6])
        instrument.client_type.natural.sell.volume = int(data[7])

    async def subscribe(self) -> None:
        """Subscribe to the channels for the appointed instruemtns"""
        self._LOGGER.info(
            "Client is subscribing to data for %d instruments.",
            len(self.subscribed_instruments),
        )
        isins = ",".join([x.identification.isin for x in self.subscribed_instruments])
        await self.websocket.send(f"1.all.{isins}")

    async def start_operation(self) -> None:
        """Start connecting to the websocket and listening for updates for a single loop"""
        self._LOGGER.info("Client is starting its single-try operation.")
        self.operation_flag = True
        await self.__operation_single_try()

    async def infinite_operation(self) -> None:
        """Start connecting to the websocket and listening for updates in an infinite loop"""
        self._LOGGER.info("Client is starting its infinite operation.")
        self.operation_flag = True
        while self.operation_flag:
            try:
                await self.__operation_single_try()
            except (OSError, ConnectionClosedError) as exc:
                print(type(Exception))
                self._LOGGER.error("Connection error: %s", repr(exc))
                await asyncio.sleep(self._OPERATION_RECONNECT_WAIT)

    async def __operation_single_try(self):
        """Does a single try on connecting to server"""
        self._LOGGER.info("Client is connecting.")
        async with client.connect(
            f"ws://{self.websocket_host}:{self.websocket_port}"
        ) as self.websocket:
            self._LOGGER.info("Client is connected.")
            await self.subscribe()
            await self.listen()

    def stop_operation(self) -> None:
        """Stops the infinite loop for operations"""
        self.operation_flag = False



"""Building the library"""
import setuptools

setuptools.setup(
    name="tsetmc_pusher",
    version="1.0.5",
    author="Arka Equities & Securities",
    author_email="zare@arkaequities.com",
    description="Pusher for Tehran Stock Exchange data crawled from TSETMC website.",
    long_description="""\
This project uses the websocket technology to build a pusher around the realtime data \
for Tehran Stock Exchange. Source data is crawled from the TSETMC website on optimized \
intervals. Users can connect to the websocket server and subscribe to as many instruments \
as they like. Afterwards, each time data for the subscribed instruments is updated, the \
changes will be pushed to the clients.
""",
    packages=setuptools.find_packages(),
    install_requires=["httpx", "websockets", "python-dotenv", "tse-utils"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)
